#!/usr/bin/env python

"""
RSVP instruction classes.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.inspect import Const, Instance

class Instruction:

    "A generic instruction."

    stack_usage = 0

    def __init__(self, attr=None):
        self.attr = attr

    def __repr__(self):
        if self.attr is not None:
            return "%s(%r)" % (self.__class__.__name__, self.attr)
        else:
            return "%s()" % self.__class__.__name__

class StackRelativeInstruction(Instruction):

    "An instruction operating on the local value stack."

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.attr.position)

SR = StackRelativeInstruction

class AddressRelativeInstruction(Instruction):

    "An instruction accessing an object's attribute."

    def __repr__(self):
        position = self.attr.position
        if isinstance(self.attr.parent, Instance):
            location = "instance"
            result = position
        else:
            location = self.attr.parent.location

            # NOTE: Unpositioned attributes are handled here.

            if location is not None and position is not None:
                result = location + position + 1
            else:
                location = self.attr.parent.name
                position = self.attr.name
                result = None
        return "%s(%r, %r -> %r)" % (self.__class__.__name__, location, position, result)

AR = AddressRelativeInstruction

class AddressInstruction(Instruction):

    "An instruction loading an address directly."

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.attr.location)

Address = AddressInstruction

class ImmediateInstruction(Instruction):

    "An instruction employing a constant."

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.attr)

Immediate = ImmediateInstruction

# Mix-in classes for stack effects.

class StackAdd:

    """
    Indicate that the stack must grow to accommodate the result of this
    instruction.
    """

    stack_usage = 1

class StackRemove:

    "Indicate that the stack must shrink as an effect of this instruction."

    stack_usage = -1

class StackRemove2:

    "Indicate that the stack must shrink as an effect of this instruction."

    stack_usage = -2

# Instructions operating on the value stack.

class LoadConst(StackAdd, Address):             "Load the constant from the specified location."
class Duplicate(StackAdd, Instruction):         "Duplicate the top of the stack."
class Pop(StackRemove, Immediate):              "Pop entries from the top of the stack."

# Access within an invocation frame.

class LoadName(StackAdd, SR):                   "Load the object from the given local attribute/variable."
class StoreName(StackRemove, SR):               "Store the object in the given local attribute/variable."
class LoadTemp(Immediate):                      "Load the object from the given temporary location."
class StoreTemp(Immediate):                     "Store the object in the given temporary location."

# Access to address-relative data.

class MakeObject(StackAdd, Instruction):        "Make a new object."
# ... DropObject not defined:                   Assume garbage collection.
class LoadAttr(AR):                             "Load the object from the given attribute."
class StoreAttr(StackRemove2, AR):              "Store an object in the given attribute."
class LoadAttrIndex(Immediate):                 "Load the object for the attribute with the given index."
class StoreAttrIndex(StackRemove2, Immediate):  "Store an object in the attribute with the given index."
class LoadAddress(StackAdd, AR):                "Load the object from the given fixed attribute address."
class StoreAddress(StackRemove, AR):            "Store an object in the given fixed attribute address."

# Access to invocation frames in preparation.

class MakeFrame(Instruction):                   "Make a new invocation frame."
class ReserveFrame(Immediate):                  "Reserve the given number of entries for the invocation frame."
class DropFrame(Instruction):                   "Drop an invocation frame."
class StoreFrame(StackRemove, Instruction):     "Store an argument at the given frame location."
class StoreFrameIndex(StackRemove, Immediate):  "Store an argument for the parameter with the given index."
class CheckFrame(Instruction):                  "Check the invocation frame for the target."
class JumpWithFrame(Instruction):               "Jump, adopting the invocation frame."

# Invocation-related instructions.

class Jump(Instruction):                        "Jump unconditionally."
class JumpIfFalse(Instruction):                 "Jump if the last evaluation gave a false result."
class JumpIfTrue(Instruction):                  "Jump if the last evaluation gave a true result."
class LoadCallable(Instruction):                "Load the target of an invocation."
class LoadContext(Instruction):                 "Load the context of an invocation."
class CheckContext(Instruction):                """Check the context of an invocation against the target,
                                                potentially discarding the context."""
class RaiseException(Instruction):              "Raise an exception."
class Return(Instruction):                      "Return a value from a subprogram."
class CheckException(Instruction):              "Check the raised exception against another."

# General instructions.

class TestIdentity(Instruction):                "Test whether the two topmost stack values are identical."
class TestIdentityAddress(Address):             "Test whether the topmost stack value is identical to the given address."

# vim: tabstop=4 expandtab shiftwidth=4
